"""Layer related George functions and enums."""

from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path
from typing import Any

from pytvpaint.george.client import send_cmd, try_cmd
from pytvpaint.george.client.parse import (
    args_dict_to_list,
    tv_cast_to_type,
    tv_parse_list,
)
from pytvpaint.george.exceptions import NoObjectWithIdError
from pytvpaint.george.grg_base import (
    BlendingMode,
    GrgErrorValue,
    RGBColor,
)


class LayerColorAction(Enum):
    """`tv_layercolor` actions.

    Attributes:
        GETCOLOR:
        SETCOLOR:
        GET:
        SET:
        LOCK:
        UNLOCK:
        SHOW:
        HIDE:
        VISIBLE:
        SELECT:
        UNSELECT:
    """

    GETCOLOR = "getcolor"
    SETCOLOR = "setcolor"
    GET = "get"
    SET = "set"
    LOCK = "lock"
    UNLOCK = "unlock"
    SHOW = "show"
    HIDE = "hide"
    VISIBLE = "visible"
    SELECT = "select"
    UNSELECT = "unselect"


class LayerColorDisplayOpt(Enum):
    """`tv_layercolorshow` display options.

    Attributes:
        DISPLAY: Activate the layers to show them in the display
        TIMELINE: Uncollpase layers from maximum collapse (2px height) in the timeline
    """

    DISPLAY = "display"
    TIMELINE = "timeline"


class InstanceNamingMode(Enum):
    """`tv_instancename` naming modes.

    Attributes:
        ALL:
        SMART:
    """

    ALL = "all"
    SMART = "smart"


class InstanceNamingProcess(Enum):
    """`tv_instancename` naming process.

    Attributes:
        EMPTY:
        NUMBER:
        TEXT:
    """

    EMPTY = "empty"
    NUMBER = "number"
    TEXT = "text"


class LayerType(Enum):
    """All the layer types.

    Attributes:
        IMAGE:
        SEQUENCE:
        XSHEET:
        SCRIBBLES:
    """

    IMAGE = "image"
    SEQUENCE = "sequence"
    XSHEET = "xsheet"
    SCRIBBLES = "scribbles"


class StencilMode(Enum):
    """All the stencil modes.

    Attributes:
        ON:
        OFF:
        NORMAL:
        INVERT:
    """

    ON = "on"
    OFF = "off"
    NORMAL = "normal"
    INVERT = "invert"


class LayerBehavior(Enum):
    """Layer behaviors on boundaries.

    Attributes:
        NONE:
        REPEAT:
        PINGPONG:
        HOLD:
    """

    NONE = "none"
    REPEAT = "repeat"
    PINGPONG = "pingpong"
    HOLD = "hold"


class LayerTransparency(Enum):
    """Layer transparency values.

    Attributes:
        ON:
        OFF:
        MINUS_1:
        NONE:
    """

    ON = "on"
    OFF = "off"
    MINUS_1 = "-1"
    NONE = "none"


class InsertDirection(Enum):
    """Instance insert direction.

    Attributes:
        BEFORE:
        AFTER:
    """

    BEFORE = "before"
    AFTER = "after"


@dataclass(frozen=True)
class TVPClipLayerColor:
    """Clip layer color values."""

    clip_id: int
    color_index: int
    color_r: int
    color_g: int
    color_b: int
    name: str


@dataclass(frozen=True)
class TVPLayer:
    """TVPaint layer info values."""

    id: int = field(metadata={"parsed": False})

    visibility: bool
    position: int
    density: int
    name: str
    type: LayerType
    first_frame: int
    last_frame: int
    selected: bool
    editable: bool
    stencil_state: StencilMode


def tv_layer_current_id() -> int:
    """Get the id of the current layer."""
    return int(send_cmd("tv_LayerCurrentId"))


@try_cmd(exception_msg="No layer at provided position")
def tv_layer_get_id(position: int) -> int:
    """Get the id of the layer at the given position.

    Raises:
        GeorgeError: if no layer found at the provided position
    """
    result = send_cmd("tv_LayerGetID", position, error_values=[GrgErrorValue.NONE])
    return int(result)


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid layer id",
)
def tv_layer_get_pos(layer_id: int) -> int:
    """Get the position of the given layer.

    Raises:
        NoObjectWithIdError: if given an invalid layer id
    """
    return int(send_cmd("tv_LayerGetPos", layer_id, error_values=[GrgErrorValue.NONE]))


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid layer id",
)
def tv_layer_info(layer_id: int) -> TVPLayer:
    """Get information of the given layer.

    Raises:
        NoObjectWithIdError: if given an invalid layer id
    """
    result = send_cmd("tv_LayerInfo", layer_id, error_values=[GrgErrorValue.EMPTY])
    layer = tv_parse_list(result, with_fields=TVPLayer, unused_indices=[7, 8])
    layer["id"] = layer_id
    return TVPLayer(**layer)


@try_cmd(exception_msg="Couldn't move current layer to position")
def tv_layer_move(position: int) -> None:
    """Move the current layer to a new position in the layer stack.

    Raises:
        GeorgeError: if layer could not be moved
    """
    send_cmd("tv_LayerMove", position)


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid layer id",
)
def tv_layer_set(layer_id: int) -> None:
    """Make the given layer the current one.

    Raises:
        NoObjectWithIdError: if given an invalid layer id
    """
    send_cmd("tv_LayerSet", layer_id)


@try_cmd(raise_exc=NoObjectWithIdError, exception_msg="Invalid layer id")
def tv_layer_selection_get(layer_id: int) -> bool:
    """Get the selection state of a layer.

    Raises:
        NoObjectWithIdError: if given an invalid layer id
    """
    res = send_cmd("tv_LayerSelection", layer_id, error_values=[-1])
    return tv_cast_to_type(res, bool)


@try_cmd(raise_exc=NoObjectWithIdError, exception_msg="Invalid layer id")
def tv_layer_selection_set(layer_id: int, new_state: bool) -> None:
    """Set the selection state of a layer.

    Raises:
        NoObjectWithIdError: if given an invalid layer id
    """
    send_cmd("tv_LayerSelection", layer_id, int(new_state), error_values=[-1])


def tv_layer_select(start_frame: int, frame_count: int) -> int:
    """Select frames in the current layer.

    Args:
        start_frame: selection start
        frame_count: number of frames to select

    Returns:
        int: number of frames selected

    Note:
        If the start position is before the beginning of the layer, the selection will only start at the beginning of
        the layer, but its length will be measured from the start position.
        This means that if you ask for a selection of 15 frames starting from position 0 in a layer that actually
        starts at position 5, only the first 10 frames in the layer will be selected.
        If the selection goes beyond the end of the layer, it will only include the frames between the start and end of
        the layer. No frames will be selected if the start position is beyond the end of the layer
    """
    return int(send_cmd("tv_LayerSelect", start_frame, frame_count, error_values=[-1]))


def tv_layer_select_info(full: bool = False) -> tuple[int, int]:
    """Get Selected frames in a layer.

    Args:
        full:  Always get the selection range, even on a non anim/ctg layer

    Returns:
        frame: the start frame of the selection
        count: the number of frames in the selection

    Bug:
        The official documentation states that this functions selects the layer frames, it does not, it simply
        returns the frames selected. This will also return all frames in the layer even if they are not selected if the
        argument `full` is set to True. We advise using `tv_layer_select` to select your frames and only using this
        function to get the selected frames.
    """
    args = ["full"] if full else []
    res = send_cmd("tv_layerSelectInfo", *args)
    frame, count = tuple(map(int, res.split(" ")))
    return frame, count


def tv_layer_create(name: str) -> int:
    """Create a new image layer with the given name."""
    return int(send_cmd("tv_LayerCreate", name, handle_string=False))


def tv_layer_duplicate(name: str) -> int:
    """Duplicate the current layer and make it the current one."""
    return int(send_cmd("tv_LayerDuplicate", name, handle_string=False))


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid layer id",
)
def tv_layer_rename(layer_id: int, name: str) -> None:
    """Rename a layer.

    Raises:
        NoObjectWithIdError: if given an invalid layer id
    """
    send_cmd("tv_LayerRename", layer_id, name)


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid layer id",
)
def tv_layer_kill(layer_id: int) -> None:
    """Delete the layer with provided id.

    Raises:
        NoObjectWithIdError: if given an invalid layer id
    """
    send_cmd("tv_LayerKill", layer_id)


def tv_layer_density_get() -> int:
    """Get the current layer density (opacity)."""
    return int(send_cmd("tv_LayerDensity"))


def tv_layer_density_set(new_density: int) -> None:
    """Set the current layer density (opacity ranging from 0 to 100)."""
    send_cmd("tv_LayerDensity", new_density)


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid layer id",
)
def tv_layer_display_get(layer_id: int) -> bool:
    """Get the visibility of the given layer.

    Raises:
        NoObjectWithIdError: if given an invalid layer id
    """
    res = send_cmd("tv_LayerDisplay", layer_id, error_values=[0])
    return tv_cast_to_type(res.lower(), bool)


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid layer id",
)
def tv_layer_display_set(
    layer_id: int, new_state: bool, light_table: bool = False
) -> None:
    """Set the visibility of the given layer.

    Raises:
        NoObjectWithIdError: if given an invalid layer id
    """
    args: list[Any] = [layer_id, int(new_state)]
    if light_table:
        args.insert(1, "lighttable")
    send_cmd("tv_LayerDisplay", *args, error_values=[0])


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid layer id",
)
def tv_layer_lock_get(layer_id: int) -> bool:
    """Get the lock state of the given layer.

    Raises:
        NoObjectWithIdError: if given an invalid layer id
    """
    res = send_cmd("tv_LayerLock", layer_id, error_values=[GrgErrorValue.ERROR])
    return tv_cast_to_type(res.lower(), bool)


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid layer id",
)
def tv_layer_lock_set(layer_id: int, new_state: bool) -> None:
    """Set the lock state of the given layer.

    Raises:
        NoObjectWithIdError: if given an invalid layer id
    """
    send_cmd(
        "tv_LayerLock",
        layer_id,
        int(new_state),
        error_values=[GrgErrorValue.ERROR],
    )


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid layer id",
)
def tv_layer_collapse_get(layer_id: int) -> bool:
    """Get the collapse mode of the given layer.

    Raises:
        NoObjectWithIdError: if given an invalid layer id
    """
    return bool(int(send_cmd("tv_LayerCollapse", layer_id, error_values=[-2])))


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid layer id",
)
def tv_layer_collapse_set(layer_id: int, new_state: bool) -> None:
    """Set the collapse mode of the given layer.

    Raises:
        NoObjectWithIdError: if given an invalid layer id
    """
    send_cmd("tv_LayerCollapse", layer_id, int(new_state), error_values=[-2])


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid layer id",
)
def tv_layer_blending_mode_get(layer_id: int) -> BlendingMode:
    """Get the blending mode of the given layer.

    Raises:
        NoObjectWithIdError: if given an invalid layer id
    """
    res = send_cmd("tv_LayerBlendingMode", layer_id)
    return tv_cast_to_type(res.lower(), BlendingMode)


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid layer id",
)
def tv_layer_blending_mode_set(layer_id: int, mode: BlendingMode) -> None:
    """Set the blending mode of the given layer.

    Raises:
        NoObjectWithIdError: if given an invalid layer id
    """
    send_cmd("tv_LayerBlendingMode", layer_id, mode.value)


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid layer id",
)
def tv_layer_stencil_get(layer_id: int) -> StencilMode:
    """Get the stencil state and mode of the given layer.

    Raises:
        NoObjectWithIdError: if given an invalid layer id
    """
    res = send_cmd("tv_LayerStencil", layer_id)
    _, state, mode = res.split(" ")

    if state == "off":
        return StencilMode.OFF

    return tv_cast_to_type(mode, StencilMode)


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid layer id",
)
def tv_layer_stencil_set(layer_id: int, mode: StencilMode) -> None:
    """Set the stencil state and mode of the given layer.

    Raises:
        NoObjectWithIdError: if given an invalid layer id
    """
    if mode == StencilMode.OFF:
        args = ["off"]
    elif mode == StencilMode.ON:
        args = ["on"]
    else:
        args = ["on", mode.value]

    send_cmd("tv_LayerStencil", layer_id, *args)


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid layer id",
)
def tv_layer_show_thumbnails_get(
    layer_id: int,
) -> bool:
    """Get the show thumbnails state for a layer.

    Raises:
        NoObjectWithIdError: if given an invalid layer id
    """
    res = send_cmd(
        "tv_LayerShowThumbnails", layer_id, error_values=[GrgErrorValue.ERROR]
    )
    return res == "1"


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid layer id",
)
def tv_layer_show_thumbnails_set(
    layer_id: int,
    state: bool,
) -> None:
    """Set the show thumbnail state for a layer.

    Raises:
        NoObjectWithIdError: if given an invalid layer id
    """
    send_cmd(
        "tv_LayerShowThumbnails",
        layer_id,
        int(state),
        error_values=[GrgErrorValue.ERROR],
    )


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid layer id",
)
def tv_layer_auto_break_instance_get(layer_id: int) -> bool:
    """Get the layer auto break instance value for a layer.

    Raises:
        NoObjectWithIdError: if given an invalid layer id
    """
    res = send_cmd("tv_LayerAutoBreakInstance", layer_id, error_values=[-1, -2, -3])
    return res == "1"


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid layer id",
)
def tv_layer_auto_break_instance_set(
    layer_id: int,
    state: bool,
) -> None:
    """Set the layer auto break instance value for a layer.

    Raises:
        NoObjectWithIdError: if given an invalid layer id
    """
    send_cmd(
        "tv_LayerAutoBreakInstance",
        layer_id,
        int(state),
        error_values=[-1, -2, -3],
    )


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid layer id",
)
def tv_layer_auto_create_instance_get(
    layer_id: int,
) -> bool:
    """Get the layer auto create instance value for a layer.

    Raises:
        NoObjectWithIdError: if given an invalid layer id
    """
    res = send_cmd("tv_LayerAutoCreateInstance", layer_id, error_values=[-1, -2, -3])
    return res == "1"


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid layer id",
)
def tv_layer_auto_create_instance_set(
    layer_id: int,
    state: bool,
) -> None:
    """Set the layer auto create instance value for a layer.

    Raises:
        NoObjectWithIdError: if given an invalid layer id
    """
    send_cmd(
        "tv_LayerAutoCreateInstance", layer_id, int(state), error_values=[-1, -2, -3]
    )


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid layer id",
)
def tv_layer_pre_behavior_get(layer_id: int) -> LayerBehavior:
    """Get the pre-behavior value for a layer.

    Raises:
        NoObjectWithIdError: if given an invalid layer id
    """
    res = send_cmd("tv_LayerPreBehavior", layer_id)
    return tv_cast_to_type(res, LayerBehavior)


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid layer id",
)
def tv_layer_pre_behavior_set(layer_id: int, behavior: LayerBehavior) -> None:
    """Set the pre-behavior value for a layer.

    Raises:
        NoObjectWithIdError: if given an invalid layer id
    """
    send_cmd("tv_LayerPreBehavior", layer_id, behavior.value)


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid layer id",
)
def tv_layer_post_behavior_get(layer_id: int) -> LayerBehavior:
    """Get the post-behavior value for a layer.

    Raises:
        NoObjectWithIdError: if given an invalid layer id
    """
    res = send_cmd("tv_LayerPostBehavior", layer_id)
    return tv_cast_to_type(res, LayerBehavior)


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid layer id",
)
def tv_layer_post_behavior_set(layer_id: int, behavior: LayerBehavior) -> None:
    """Set the post-behavior value for a layer.

    Raises:
        NoObjectWithIdError: if given an invalid layer id
    """
    send_cmd("tv_LayerPostBehavior", layer_id, behavior.value)


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid layer id",
)
def tv_layer_lock_position_get(layer_id: int) -> bool:
    """Get the lock position state of a layer.

    Raises:
        NoObjectWithIdError: if given an invalid layer id
    """
    res = send_cmd("tv_LayerLockPosition", layer_id)
    return tv_cast_to_type(res, bool)


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid layer id",
)
def tv_layer_lock_position_set(layer_id: int, state: bool) -> None:
    """Set the lock position state of a layer.

    Raises:
        NoObjectWithIdError: if given an invalid layer id
    """
    send_cmd("tv_LayerLockPosition", layer_id, int(state))


def tv_preserve_get() -> LayerTransparency:
    """Get the preserve transparency state of the current layer."""
    res = send_cmd("tv_Preserve")
    _, state = res.split(" ")
    return tv_cast_to_type(state, LayerTransparency)


def tv_preserve_set(state: LayerTransparency) -> None:
    """Set the preserve transparency state of the current layer."""
    send_cmd("tv_Preserve", "alpha", state.value)


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid layer id",
)
def tv_layer_mark_get(layer_id: int, frame: int) -> int:
    """Get the mark color of a layer at a frame.

    Args:
        layer_id: the layer id
        frame: the frame with a mark

    Raises:
        NoObjectWithIdError: if given an invalid layer id

    Returns:
        int: the mark color index
    """
    return int(send_cmd("tv_LayerMarkGet", layer_id, frame))


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid layer id",
)
def tv_layer_mark_set(layer_id: int, frame: int, color_index: int) -> None:
    """Set the mark of the layer's frame.

    Args:
        layer_id: the layer id
        frame: the frame to set the mark (use 0 to remove it).
        color_index: the mark color

    Raises:
        NoObjectWithIdError: if given an invalid layer id
    """
    send_cmd("tv_LayerMarkSet", layer_id, frame, color_index)


def tv_layer_anim(layer_id: int) -> None:
    """Convert the layer to an anim layer."""
    send_cmd("tv_LayerAnim", *([layer_id] if layer_id else []))


def tv_layer_copy() -> None:
    """Copy the current image or the selected ones."""
    send_cmd("tv_LayerCopy")


def tv_layer_cut() -> None:
    """Cut the current image or the selected ones."""
    send_cmd("tv_LayerCut")


def tv_layer_paste() -> None:
    """Paste the previously copied/cut images to the current layer."""
    send_cmd("tv_LayerPaste")


def tv_layer_insert_image(
    count: int = 1,
    direction: InsertDirection | None = None,
    duplicate: bool | None = None,
) -> None:
    """Add new image(s) before/after the current one and make it current."""
    if duplicate:
        args = [0]
    else:
        args_dict = {
            "count": count,
            "direction": direction.value if direction is not None else None,
        }
        args = args_dict_to_list(args_dict)

    send_cmd("tv_LayerInsertImage", *args)


def tv_layer_merge(
    layer_id: int,
    blending_mode: BlendingMode,
    stamp: bool = False,
    erase: bool = False,
    keep_color_grp: bool = True,
    keep_img_mark: bool = True,
    keep_instance_name: bool = True,
) -> None:
    """Merge the given layer with the current one.

    Args:
        layer_id: the layer id
        blending_mode: the blending mode to use
        stamp: Use stamp mode
        erase: Remove the source layer
        keep_color_grp: Keep the color group
        keep_img_mark: Keep the image mark
        keep_instance_name: Keep the instance name
    """
    args = [
        layer_id,
        blending_mode.value,
    ]

    if stamp:
        args.append("stamp")
    if erase:
        args.append("erase")

    args_dict = {
        "keepcolorgroup": int(keep_color_grp),
        "keepimagemark": int(keep_img_mark),
        "keepinstancename": int(keep_instance_name),
    }
    args.extend(args_dict_to_list(args_dict))

    send_cmd("tv_LayerMerge", layer_id, *args)


def tv_layer_merge_all(
    keep_color_grp: bool = True,
    keep_img_mark: bool = True,
    keep_instance_name: bool = True,
) -> None:
    """Merge all layers.

    Args:
        keep_color_grp: Keep the color group
        keep_img_mark: Keep the image mark
        keep_instance_name: Keep the instance name
    """
    args_dict = {
        "keepcolorgroup": int(keep_color_grp),
        "keepimagemark": int(keep_img_mark),
        "keepinstancename": int(keep_instance_name),
    }
    send_cmd("tv_LayerMergeAll", *args_dict_to_list(args_dict))


def tv_layer_shift(layer_id: int, start: int) -> None:
    """Move the layer to a new frame.

    Args:
        layer_id: layer id
        start: frame to shift layer to
    """
    send_cmd("tv_LayerShift", layer_id, start)


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid layer id",
)
def tv_layer_load_dependencies(layer_id: int) -> None:
    """Load all dependencies of the given layer in memory.

    Raises:
        NoObjectWithIdError: if given an invalid layer id
    """
    send_cmd("tv_LayerLoadDependencies", layer_id)


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid layer id",
)
def tv_layer_color_get_color(clip_id: int, color_index: int) -> TVPClipLayerColor:
    """Get a specific colors information in the clips color list.

    Raises:
        NoObjectWithIdError: if given an invalid layer id
    """
    result = send_cmd(
        "tv_LayerColor",
        LayerColorAction.GETCOLOR.value,
        clip_id,
        color_index,
        error_values=[GrgErrorValue.ERROR],
    )
    parsed = tv_parse_list(result, with_fields=TVPClipLayerColor)
    return TVPClipLayerColor(**parsed)


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid layer id",
)
def tv_layer_color_set_color(
    clip_id: int,
    color_index: int,
    color: RGBColor,
    name: str | None = None,
) -> None:
    """Set a specific colors information in the clips color list.

    Raises:
        NoObjectWithIdError: if given an invalid layer id

    Note:
        The color with index 0 is the "Default" color, and it can't be changed
    """
    args: list[Any] = [
        LayerColorAction.SETCOLOR.value,
        clip_id,
        color_index,
        color.r,
        color.g,
        color.b,
    ]

    if name:
        args.append(name)

    send_cmd("tv_LayerColor", *args, error_values=[GrgErrorValue.ERROR])


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid layer id",
)
def tv_layer_color_get(layer_id: int) -> int:
    """Get the layer's color index from the clips color list.

    Raises:
        NoObjectWithIdError: if given an invalid layer id
    """
    res = send_cmd(
        "tv_LayerColor",
        LayerColorAction.GET.value,
        layer_id,
        error_values=[-1],
    )
    return int(res)


@try_cmd(raise_exc=NoObjectWithIdError)
def tv_layer_color_set(layer_id: int, color_index: int) -> None:
    """Set the layer's color index from the clips color list.

    Raises:
        NoObjectWithIdError: if given an invalid layer id
    """
    send_cmd(
        "tv_LayerColor",
        LayerColorAction.SET.value,
        layer_id,
        color_index,
        error_values=[-1],
    )


def tv_layer_color_lock(color_index: int) -> int:
    """Lock all layers that use the given color index.

    Args:
        color_index: the layer color index

    Returns:
        the number of layers locked
    """
    return int(send_cmd("tv_LayerColor", LayerColorAction.LOCK.value, color_index))


def tv_layer_color_unlock(color_index: int) -> int:
    """Unlock all layers that use the given color index.

    Args:
        color_index: the layer color index

    Returns:
        the number of unlocked layers
    """
    return int(send_cmd("tv_LayerColor", LayerColorAction.UNLOCK.value, color_index))


def tv_layer_color_show(mode: LayerColorDisplayOpt, color_index: int) -> int:
    """Show all layers that use the given color index.

    Args:
        mode: the display mode
        color_index: the layer color index

    Returns:
        the number of unlocked layers
    """
    res = send_cmd(
        "tv_LayerColor",
        LayerColorAction.SHOW.value,
        mode.value,
        color_index,
        error_values=[GrgErrorValue.ERROR],
    )
    return int(res)


def tv_layer_color_hide(mode: LayerColorDisplayOpt, color_index: int) -> int:
    """Hide all layers that use the given color index.

    Args:
        mode: the display mode
        color_index: the layer color index

    Returns:
        the number of unlocked layers
    """
    return int(
        send_cmd(
            "tv_LayerColor",
            LayerColorAction.HIDE.value,
            mode.value,
            color_index,
            error_values=[GrgErrorValue.ERROR],
        )
    )


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid layer id",
)
def tv_layer_color_visible(color_index: int) -> bool:
    """Get the visibility of the color index (2px height) in the timeline.

    Raises:
        NoObjectWithIdError: if given an invalid layer id
    """
    return bool(
        send_cmd(
            "tv_LayerColor",
            LayerColorAction.VISIBLE.value,
            color_index,
            error_values=[-1],
        )
    )


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid color index",
)
def tv_layer_color_select(color_index: int) -> int:
    """Select all layers that use the given color index.

    Raises:
        NoObjectWithIdError: if given an invalid layer id
    """
    return int(send_cmd("tv_LayerColor", LayerColorAction.SELECT.value, color_index))


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid color index",
)
def tv_layer_color_unselect(color_index: int) -> int:
    """Unselect all layers that use the given color index.

    Raises:
        NoObjectWithIdError: if given an invalid layer id
    """
    return int(send_cmd("tv_LayerColor", LayerColorAction.UNSELECT.value, color_index))


def tv_instance_name(
    layer_id: int,
    mode: InstanceNamingMode,
    prefix: str | None = None,
    suffix: str | None = None,
    process: InstanceNamingProcess | None = None,
) -> None:
    """Rename all instances.

    Note:
        The suffix can only be added when using mode InstanceNamingMode.SMART

    Bug:
        Using a wrong layer_id causes a crash

    Args:
        layer_id: the layer id
        mode: the instance renaming mode
        prefix: the prefix to add to each name
        suffix: the suffix to add to each name
        process: the instance naming process
    """
    args_dict: dict[str, Any] = {
        "mode": mode.value,
        "prefix": prefix,
    }

    if mode == InstanceNamingMode.SMART:
        args_dict["suffix"] = suffix
        args_dict["process"] = process.value if process else None

    args = args_dict_to_list(args_dict)
    send_cmd("tv_InstanceName", layer_id, *args, error_values=[-1, -2])


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid layer id or the frame doesn't have an instance",
)
def tv_instance_get_name(layer_id: int, frame: int) -> str:
    """Get the name of an instance.

    Args:
        layer_id: the layer id
        frame: the frame of the instance

    Raises:
        NoObjectWithIdError: if given an invalid layer id or an invalid instance frame

    Returns:
        the instance name
    """
    return send_cmd("tv_InstanceGetName", layer_id, frame).strip('"')


@try_cmd(exception_msg="Invalid layer id or no instance at given frame")
def tv_instance_set_name(layer_id: int, frame: int, name: str) -> str:
    """Set the name of an instance.

    Raises:
        GeorgeError: if an invalid layer id was provided or no instance was found at the given frame
    """
    return send_cmd("tv_InstanceSetName", layer_id, frame, name)


def tv_exposure_next() -> int:
    """Go to the next layer instance head.

    Returns:
        The next instances start frame
    """
    return int(send_cmd("tv_ExposureNext"))


def tv_exposure_break(frame: int) -> None:
    """Break a layer instance/exposure at the given frame.

    Args:
        frame: the split frame
    """
    send_cmd("tv_ExposureBreak", frame)


def tv_exposure_add(frame: int, count: int) -> None:
    """Add new frames to an existing layer instance/exposure.

    Args:
        frame: the split frame
        count: the number of frames to add
    """
    send_cmd("tv_ExposureAdd", frame, count)


def tv_exposure_set(frame: int, count: int) -> None:
    """Set the number frames of an existing layer instance/exposure.

    Args:
        frame: the split frame
        count: the number of frames to add
    """
    send_cmd("tv_ExposureSet", frame, count)


def tv_exposure_prev() -> int:
    """Go to the previous layer instance head (*before* the current instance).

    Returns:
        The previous instances start frame
    """
    return int(send_cmd("tv_ExposurePrev"))


@try_cmd(exception_msg="No file found or invalid format")
def tv_save_image(export_path: Path | str) -> None:
    """Save the current image of the current layer.

    Raises:
        GeorgeError: if the file couldn't be saved or an invalid format was provided
    """
    export_path = Path(export_path)
    send_cmd("tv_SaveImage", export_path.as_posix())


@try_cmd(exception_msg="Invalid image format")
def tv_load_image(img_path: Path | str, stretch: bool | None = None) -> None:
    """Load an image in the current image layer.

    Raises:
        FileNotFoundError: if the input file doesn't exist
        GeorgeError: if the provided file is in an invalid format
    """
    img_path = Path(img_path)

    if not img_path.exists():
        raise FileNotFoundError(f"File not found at: {img_path.as_posix()}")

    args: list[Any] = [img_path.as_posix()]
    if stretch:
        args.append("stretch")

    send_cmd("tv_LoadImage", *args)


def tv_clear(fill_b_pen: bool = False) -> None:
    """Clear (or fill with BPen) the current image (selection) of the current layer."""
    send_cmd("tv_Clear", int(fill_b_pen))



"""Scene class."""

from __future__ import annotations

from collections.abc import Iterator

from pytvpaint import george, utils
from pytvpaint.clip import Clip
from pytvpaint.project import Project
from pytvpaint.utils import (
    Removable,
    set_as_current,
)


class Scene(Removable):
    """A Scene is a collection of clips. A Scene is inside a project."""

    def __init__(self, scene_id: int, project: Project) -> None:
        super().__init__()
        self._id: int = scene_id
        self._project = project

    def __repr__(self) -> str:
        """String representation of the scene."""
        return f"Scene()<id:{self.id}>"

    def __eq__(self, other: object) -> bool:
        """Two scenes are equal if their id is the same."""
        if not isinstance(other, Scene):
            return NotImplemented
        return self.id == other.id

    @staticmethod
    def current_scene_id() -> int:
        """Returns the current scene id (the current clip's scene)."""
        return george.tv_scene_current_id()

    @staticmethod
    def current_scene() -> Scene:
        """Returns the current scene of the current project."""
        return Scene(
            scene_id=george.tv_scene_current_id(),
            project=Project.current_project(),
        )

    @classmethod
    def new(cls, project: Project | None = None) -> Scene:
        """Creates a new scene in the provided project."""
        project = project or Project.current_project()
        project.make_current()
        george.tv_scene_new()
        return cls.current_scene()

    def make_current(self) -> None:
        """Make this scene the current one."""
        if self.is_current:
            return

        # In order to select the scene, we select any child clip inside of it
        first_clip_id = george.tv_clip_enum_id(self.id, 0)
        george.tv_clip_select(first_clip_id)

    @property
    def is_current(self) -> bool:
        """Returns `True` if the scene is the current one."""
        return self.id == self.current_scene_id()

    @property
    def id(self) -> int:
        """The scene id."""
        return self._id

    @property
    def project(self) -> Project:
        """The scene's project."""
        return self._project

    @property
    def position(self) -> int:
        """The scene's position in the project.

        Raises:
            ValueError: if scene cannot be found in the project
        """
        for pos, other_id in enumerate(self.project.current_scene_ids()):
            if other_id == self.id:
                return pos
        raise Exception("The scene doesn't exist anymore")

    @position.setter
    def position(self, value: int) -> None:
        value = max(0, value)
        george.tv_scene_move(self.id, value)

    @property
    @set_as_current
    def clip_ids(self) -> Iterator[int]:
        """Returns an iterator over the clip ids."""
        return utils.position_generator(
            lambda pos: george.tv_clip_enum_id(self.id, pos)
        )

    @property
    def clips(self) -> Iterator[Clip]:
        """Yields the scene clips."""
        for clip_id in self.clip_ids:
            yield Clip(clip_id, project=self._project)

    def get_clip(
        self,
        by_id: int | None = None,
        by_name: str | None = None,
    ) -> Clip | None:
        """Find a clip by id or by name."""
        for clip in self.clips:
            if (by_id and clip.id == by_id) or (by_name and clip.name == by_name):
                return clip

        return None

    @set_as_current
    def add_clip(self, clip_name: str) -> Clip:
        """Adds a new clip to the scene."""
        self.make_current()
        return Clip.new(name=clip_name, project=self.project)

    def duplicate(self) -> Scene:
        """Duplicate the scene and return it."""
        self.project.make_current()
        george.tv_scene_duplicate(self.id)
        dup_pos = self.position + 1
        dup_id = george.tv_scene_enum_id(dup_pos)
        return Scene(dup_id, self.project)

    def remove(self) -> None:
        """Remove the scene and all the clips inside.

        Warning:
            All `Clip` instances will be invalid after removing the scene.
            There's no protection mechanism to prevent accessing clip data that doesn't exist anymore.
        """
        george.tv_scene_close(self._id)
        self.mark_removed()



"""JSON-RPC client and data models."""

from __future__ import annotations

import contextlib
import json
import sys
import threading
from time import time
from typing import Any, Union, cast

from typing_extensions import NotRequired, TypedDict
from websocket import WebSocket, WebSocketException

from pytvpaint import log

JSONValueType = Union[str, int, float, bool, None]


class JSONRPCPayload(TypedDict):
    """A rpc call is represented by sending a Request object to a Server.

    See: https://www.jsonrpc.org/specification#request_object
    """

    jsonrpc: str
    id: int
    method: str
    params: list[JSONValueType]


class JSONRPCResponse(TypedDict):
    """When a rpc call is made, the Server MUST reply with a Response.

    See: https://www.jsonrpc.org/specification#response_object
    """

    jsonrpc: str
    id: int
    result: str
    error: NotRequired[JSONRPCError]


class JSONRPCError(TypedDict):
    """When a rpc call encounters an error.

    See: https://www.jsonrpc.org/specification#error_object
    """

    code: int
    message: str
    data: NotRequired[Any]


class JSONRPCResponseError(Exception):
    """Exception used when a rpc call encounters an error."""

    def __init__(self, error: JSONRPCError) -> None:
        super().__init__(f"JSON-RPC Server error ({error['code']}): {error['message']}")


class JSONRPCClient:
    """Simple JSON-RPC 2.0 client over websockets with automatic reconnection.

    See: https://www.jsonrpc.org/specification#notification
    """

    def __init__(self, url: str, timeout: int = 60, version: str = "2.0") -> None:
        """Initialize a new JSON-RPC client with a WebSocket url endpoint.

        Args:
            url: the WebSocket url endpoint
            timeout: the reconnection timeout
            version: The JSON-RPC version. Defaults to "2.0".
        """
        self.ws_handle = WebSocket()
        self.url = url
        self.rpc_id = 0
        self.timeout = timeout
        self.jsonrpc_version = version

        self.stop_ping = threading.Event()
        self.run_forever = False
        self.ping_thread: threading.Thread | None = None
        self._ping_start_time: float = 0

    def _auto_reconnect(self) -> None:
        """Automatic WebSocket reconnection in a thread by pinging the server."""
        while self.run_forever and not self.stop_ping.wait(1):
            try:
                self.ws_handle.ping()
                continue
            except (WebSocketException, ConnectionError):
                self.ws_handle.close()

            with contextlib.suppress(ConnectionRefusedError):
                self.connect()
                log.info(f"Reconnected automatically to endpoint: {self.url}")
                continue

            # There's a timeout after which we stop reconnecting
            if self.timeout and (time() - self._ping_start_time) > self.timeout:
                raise ConnectionRefusedError(
                    "Could not establish connection with a tvpaint instance before timeout !"
                )

    def __del__(self) -> None:
        """Called when the client goes out of scope."""
        self.disconnect()

    @property
    def is_connected(self) -> bool:
        """Returns True if the client is connected."""
        return self.ws_handle.connected

    def connect(self, timeout: float | None = None) -> None:
        """Connects to the WebSocket endpoint."""
        self.ws_handle.connect(self.url, timeout=timeout)

        if not self.ping_thread:
            self._ping_start_time = time()
            self.ping_thread = threading.Thread(
                target=self._auto_reconnect, daemon=True
            )
            self.run_forever = True
            self.ping_thread.start()

    def disconnect(self) -> None:
        """Disconnects from the server."""
        self.run_forever = False
        if self.ping_thread:
            self.ping_thread.join()

        self.ws_handle.close()

    def increment_rpc_id(self) -> None:
        """Increments the internal RPC id until it reaches `sys.maxsize`."""
        self.rpc_id = (self.rpc_id + 1) % sys.maxsize

    def execute_remote(
        self,
        method: str,
        params: list[JSONValueType] | None = None,
    ) -> JSONRPCResponse:
        """Executes a remote procedure call.

        Args:
            method: the name of the method to be invoked
            params: the parameter values to be used during the invocation of the method. Defaults to None.

        Raises:
            ConnectionError: if the client is not connected
            JSONRPCResponseError: if there was an error server-side

        Returns:
            JSONRPCResponse: the JSON-RPC response payload
        """
        if not self.is_connected:
            raise ConnectionError(
                f"Can't send rpc message because the client is not connected to {self.url}"
            )

        payload: JSONRPCPayload = {
            "jsonrpc": self.jsonrpc_version,
            "id": self.rpc_id,
            "method": method,
            "params": params or [],
        }

        self.ws_handle.send(json.dumps(payload))
        self.increment_rpc_id()

        result = self.ws_handle.recv()
        response = cast(JSONRPCResponse, json.loads(result))

        if "error" in response:
            raise JSONRPCResponseError(response["error"])

        return response



"""This modules handles the WebSocket client and automatically connect to the server running in TVPaint.

It also has crucial functions like `send_cmd` that send George commands and get the result.
"""

from __future__ import annotations

import contextlib
import functools
import os
import re
from pathlib import Path
from time import sleep, time
from typing import Any, Callable, TypeVar, cast

from pytvpaint import log
from pytvpaint.george.client.parse import tv_handle_string
from pytvpaint.george.client.rpc import JSONRPCClient
from pytvpaint.george.exceptions import GeorgeError


def _connect_client(
    host: str = "ws://localhost", port: int = 3000, timeout: int = 60
) -> JSONRPCClient:
    host = os.getenv("PYTVPAINT_WS_HOST", host)
    port = int(os.getenv("PYTVPAINT_WS_PORT", port))
    startup_connect = bool(int(os.getenv("PYTVPAINT_WS_STARTUP_CONNECT", 1)))
    timeout = int(os.getenv("PYTVPAINT_WS_TIMEOUT", timeout))

    rpc_client = JSONRPCClient(f"{host}:{port}", timeout)

    if not startup_connect:
        return rpc_client

    start_time = time()
    wait_duration = 5
    connection_successful = False

    while True:
        if timeout and (time() - start_time) > timeout:
            break
        with contextlib.suppress(ConnectionRefusedError):
            rpc_client.connect()
            connection_successful = True
            break

        log.warning(f"Connection refused, trying again in {wait_duration} seconds...")
        sleep(wait_duration)

    if not connection_successful:
        # Connection could not be established after timeout
        if rpc_client.is_connected:
            rpc_client.disconnect()

        raise ConnectionRefusedError(
            "Could not establish connection with a tvpaint instance before timeout !"
        )

    if connection_successful:
        log.info(f"Connected to TVPaint on port {port}")

    return rpc_client


rpc_client = _connect_client()

T = TypeVar("T", bound=Callable[..., Any])


def try_cmd(
    raise_exc: type[Exception] = GeorgeError,
    catch_exc: type[Exception] = GeorgeError,
    exception_msg: str | None = None,
) -> Callable[[T], T]:
    """Decorator that does a try/except with GeorgeError by default.

    It raises the error with the custom exception message provided.

    Args:
        raise_exc: the exception to raise. Defaults to GeorgeError.
        catch_exc: the exception to catch. Defaults to GeorgeError.
        exception_msg: Custom exception message. Defaults to None.

    Returns:
        the decorated function
    """

    def decorate(func: T) -> T:
        @functools.wraps(func)
        def applicator(*args: Any, **kwargs: Any) -> Any:
            try:
                return func(*args, **kwargs)
            except catch_exc as e:
                raise raise_exc(exception_msg or e)

        return cast(T, applicator)

    return decorate


def send_cmd(
    command: str,
    *args: Any,
    error_values: list[Any] | None = None,
    handle_string: bool = True,
) -> str:
    """Send a George command with the provided arguments to TVPaint.

    Catch basic `ERROR XX` errors returned from George, but you can provide your own error values.

    Args:
        command: the George command to send
        *args: pass any arguments you want to that function
        error_values: a list of error values to catch from George. Defaults to None.
        handle_string: control the quote wrapping of string with spaces. Defaults to True.

    Raises:
        GeorgeError: if we received `ERROR XX` or any of the custom error codes

    Returns:
        the George return string
    """
    tv_args = [
        tv_handle_string(arg) if handle_string and isinstance(arg, str) else arg
        for arg in args
    ]
    cmd_str = " ".join([str(arg) for arg in [command, *tv_args]])

    is_stack = command in ["tv_UndoOpenStack", "tv_UpdateUndo", "tv_UndoCloseStack"]

    if not is_stack:
        log.debug(f"[RPC] >> {cmd_str}")

    response = rpc_client.execute_remote("execute_george", [cmd_str])
    result = response["result"]

    if not is_stack:
        log.debug(f"[RPC] << {result}")

    # Test for basic ERROR X values and user provided custom errors
    res_in_error_values = error_values and result in list(map(str, error_values))
    if res_in_error_values or re.match(r"ERROR -?\d+", result, re.IGNORECASE):
        msg = f"Received value: '{result}' considered as an error"
        raise GeorgeError(msg, error_value=result)

    return result


def run_script(script: Path | str) -> None:
    """Execute a George script from a .grg file.

    Args:
        script: the path to the script

    Raises:
        ValueError: if the script was not found
    """
    script = Path(script)
    if not script.exists():
        raise ValueError(f"Script not found at : {script.as_posix()}")
    send_cmd("tv_RunScript", script.as_posix())



"""Project related George functions and enums."""

from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path
from typing import Any

from pytvpaint.george.client import send_cmd, try_cmd
from pytvpaint.george.client.parse import (
    tv_cast_to_type,
    tv_parse_list,
    validate_args_list,
)
from pytvpaint.george.exceptions import NoObjectWithIdError
from pytvpaint.george.grg_base import (
    FieldOrder,
    GrgErrorValue,
    ResizeOption,
    RGBColor,
    TVPSound,
)


@dataclass(frozen=True)
class TVPProject:
    """TVPaint project info values."""

    id: str = field(metadata={"parsed": False})

    path: Path
    width: int
    height: int
    pixel_aspect_ratio: float
    frame_rate: float
    field_order: FieldOrder
    start_frame: int


class BackgroundMode(Enum):
    """The project background mode.

    Attributes:
        CHECK:
        COLOR:
        NONE:
    """

    CHECK = "check"
    COLOR = "color"
    NONE = "none"


def tv_background_get() -> (
    tuple[BackgroundMode, tuple[RGBColor, RGBColor] | RGBColor | None]
):
    """Get the background mode of the project, and the color(s) if in `color` or `check` mode.

    Returns:
        mode: the background mode
        colors: the background colors if any

    """
    res = send_cmd("tv_Background")

    mode, *values = res.split(" ")

    if mode == BackgroundMode.NONE.value:
        return BackgroundMode.NONE, None

    if mode == BackgroundMode.CHECK.value:
        c1 = map(int, values[:3])
        c2 = map(int, values[3:])
        return BackgroundMode.CHECK, (RGBColor(*c1), RGBColor(*c2))

    return BackgroundMode.COLOR, RGBColor(*map(int, values))


def tv_background_set(
    mode: BackgroundMode,
    color: tuple[RGBColor, RGBColor] | RGBColor | None = None,
) -> None:
    """Set the background mode of the project.

    Args:
        mode: color mode (None, checker or one color)
        color: None for None mode, RBGColor for one color, and tuple of RGBColors for checker
    """
    args = []

    if mode == BackgroundMode.CHECK and isinstance(color, tuple):
        c1, c2 = color
        args = [c1.r, c1.g, c1.b, c2.r, c2.g, c2.b]
    elif mode == BackgroundMode.COLOR and isinstance(color, RGBColor):
        args = [color.r, color.g, color.b]

    send_cmd("tv_Background", mode.value, *args)


@try_cmd(exception_msg="Project created but may be corrupted")
def tv_project_new(
    project_path: Path | str,
    width: int = 1920,
    height: int = 1080,
    pixel_aspect_ratio: float = 1.0,
    frame_rate: float = 24.0,
    field_order: FieldOrder = FieldOrder.NONE,
    start_frame: int = 1,
) -> str:
    """Create a new project.

    Raises:
        GeorgeError: if an error occurred during the project creation
    """
    return send_cmd(
        "tv_ProjectNew",
        Path(project_path).as_posix(),
        width,
        height,
        pixel_aspect_ratio,
        frame_rate,
        field_order.value,
        start_frame,
        error_values=[GrgErrorValue.EMPTY],
    )


@try_cmd(exception_msg="Invalid format")
def tv_load_project(project_path: Path | str, silent: bool | None = None) -> str:
    """Load a file as a project if possible or open Import panel.

    Raises:
        FileNotFoundError: if the project file doesn't exist
        GeorgeError: if the provided file is in an invalid format
    """
    project_path = Path(project_path)

    if not project_path.exists():
        raise FileNotFoundError(f"Project not found at: {project_path.as_posix()}")

    args: list[Any] = [project_path.as_posix()]

    if silent is not None:
        args.extend(["silent", int(silent)])

    return send_cmd("tv_LoadProject", *args, error_values=[-1])


def tv_save_project(project_path: Path | str) -> None:
    """Save the current project as tvpp."""
    project_path = Path(project_path)
    parent = project_path.parent

    if not parent.exists():
        msg = f"Can't save because parent folder does not exist: {parent.as_posix()}"
        raise ValueError(msg)

    send_cmd("tv_SaveProject", project_path.as_posix())


@try_cmd(exception_msg="Can't duplicate the current project")
def tv_project_duplicate() -> None:
    """Duplicate the current project.

    Raises:
        GeorgeError: if an error occurred during the project creation.
    """
    send_cmd("tv_ProjectDuplicate", error_values=[0])


@try_cmd(exception_msg="No project at provided position")
def tv_project_enum_id(position: int) -> str:
    """Get the id of the project at the given position.

    Raises:
        GeorgeError: if no project found at the provided position.
    """
    return send_cmd("tv_ProjectEnumId", position, error_values=[GrgErrorValue.NONE])


def tv_project_current_id() -> str:
    """Get the id of the current project."""
    return send_cmd("tv_ProjectCurrentId")


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid project id",
)
def tv_project_info(project_id: str) -> TVPProject:
    """Get info of the given project.

    Raises:
        NoObjectWithIdError: if given an invalid project id
    """
    result = send_cmd("tv_ProjectInfo", project_id, error_values=[GrgErrorValue.EMPTY])
    project = tv_parse_list(result, with_fields=TVPProject)
    project["id"] = project_id
    return TVPProject(**project)


def tv_get_project_name() -> str:
    """Returns the save path of the current project."""
    return send_cmd("tv_GetProjectName")


def tv_project_select(project_id: str) -> str:
    """Make the given project current."""
    return send_cmd("tv_ProjectSelect", project_id)


def tv_project_close(project_id: str) -> None:
    """Close the given project."""
    send_cmd("tv_ProjectClose", project_id)


def tv_resize_project(width: int, height: int) -> None:
    """Resize the current project.

    Note:
        creates a resized copy of the project with a new id
    """
    send_cmd("tv_ResizeProject", width, height)


def tv_resize_page(width: int, height: int, resize_opt: ResizeOption) -> None:
    """Create a new resized project and close the current one."""
    send_cmd("tv_ResizePage", width, height, resize_opt.value)


def tv_get_width() -> int:
    """Get the current project width."""
    return int(send_cmd("tv_GetWidth"))


def tv_get_height() -> int:
    """Get the current project height."""
    return int(send_cmd("tv_GetHeight"))


def tv_ratio() -> float:
    """Get the current project pixel aspect ratio.

    Bug:
        Doesn't work and always returns an empty string
    """
    return float(send_cmd("tv_GetRatio", error_values=[GrgErrorValue.EMPTY]))


def tv_get_field() -> FieldOrder:
    """Get the current project field mode."""
    return tv_cast_to_type(send_cmd("tv_GetField"), cast_type=FieldOrder)


def tv_project_save_sequence(
    export_path: Path | str,
    use_camera: bool | None = None,
    start: int | None = None,
    end: int | None = None,
) -> None:
    """Save the current project."""
    export_path = Path(export_path).resolve()
    args: list[Any] = [export_path.as_posix()]

    if use_camera:
        args.append("camera")
    if start and end:
        args.extend((start, end))

    send_cmd(
        "tv_ProjectSaveSequence",
        *args,
        error_values=[-1],
    )


def tv_project_render_camera(project_id: str) -> str:
    """Render the given project's camera view to a new project.

    Returns:
        the new project id
    """
    return send_cmd(
        "tv_ProjectRenderCamera",
        project_id,
        error_values=[GrgErrorValue.ERROR],
    )


def tv_frame_rate_get() -> tuple[float, float]:
    """Get the framerate of the current project."""
    parse = tv_parse_list(
        send_cmd("tv_FrameRate", 1, "info"),
        with_fields=[
            ("project_fps", float),
            ("playback_fps", float),
        ],
    )
    project_fps, playback_fps = parse.values()
    return project_fps, playback_fps


def tv_frame_rate_set(
    frame_rate: float, time_stretch: bool = False, preview: bool = False
) -> None:
    """Get the framerate of the current project."""
    args: list[Any] = []
    if time_stretch:
        args = ["timestretch"]
    if preview:
        args = ["preview"]
    args.insert(0, frame_rate)
    send_cmd("tv_FrameRate", *args)


def tv_frame_rate_project_set(
    frame_rate: float, time_stretch: bool | None = None
) -> None:
    """Set the framerate of the current project."""
    args: list[Any] = [frame_rate]
    if time_stretch:
        args.append("timestretch")
    send_cmd("tv_FrameRate", *args)


def tv_frame_rate_preview_set(frame_rate: float) -> None:
    """Set the framerate of the preview (playback)."""
    send_cmd("tv_FrameRate", frame_rate, "preview")


def tv_project_current_frame_get() -> int:
    """Get the current frame of the current project."""
    return int(send_cmd("tv_ProjectCurrentFrame"))


def tv_project_current_frame_set(frame: int) -> int:
    """Set the current frame of the current project.

    Note:
        this is relative to the current clip markin
    """
    return int(send_cmd("tv_ProjectCurrentFrame", frame))


def tv_load_palette(palette_path: Path | str) -> None:
    """Load a palette(s) from a file/directory.

    Raises:
        FileNotFoundError: if palette was not found at the provided path
    """
    palette_path = Path(palette_path)
    if not palette_path.exists():
        raise FileNotFoundError(f"Palette not found at: {palette_path.as_posix()}")
    send_cmd("tv_LoadPalette", palette_path.as_posix())


def tv_save_palette(palette_path: Path | str) -> None:
    """Save the current palette.

    Raises:
        FileNotFoundError: if palette save directory doesn't exist
    """
    palette_path = Path(palette_path)

    if not palette_path.parent.exists():
        parent_path = palette_path.parent.as_posix()
        raise NotADirectoryError(
            f"Can't save palette because parent folder doesn't exist: {parent_path}"
        )

    send_cmd("tv_SavePalette", palette_path.as_posix())


def tv_project_save_video_dependencies() -> None:
    """Saves current project video dependencies."""
    send_cmd("tv_ProjectSaveVideoDependencies")


def tv_project_save_audio_dependencies() -> None:
    """Saves current project audio dependencies."""
    send_cmd("tv_ProjectSaveAudioDependencies")


def tv_sound_project_info(project_id: str, track_index: int) -> TVPSound:
    """Get information about a project soundtrack."""
    res = send_cmd(
        "tv_SoundProjectInfo", project_id, track_index, error_values=[-1, -2, -3]
    )
    res_parse = tv_parse_list(res, with_fields=TVPSound)
    return TVPSound(**res_parse)


def tv_sound_project_new(sound_path: Path | str) -> None:
    """Add a new soundtrack to the current project."""
    path = Path(sound_path)
    if not path.exists():
        raise ValueError(f"Sound file not found at : {path.as_posix()}")

    send_cmd("tv_SoundProjectNew", path.as_posix(), error_values=[-1, -3, -4])


def tv_sound_project_remove(track_index: int) -> None:
    """Remove a soundtrack from the current project."""
    send_cmd("tv_SoundProjectRemove", track_index, error_values=[-2])


def tv_sound_project_reload(project_id: str, track_index: int) -> None:
    """Reload a project soundtracks file."""
    send_cmd(
        "tv_SoundProjectReload",
        project_id,
        track_index,
        error_values=[-1, -2, -3],
    )


def tv_sound_project_adjust(
    track_index: int,
    mute: bool | None = None,
    volume: float | None = None,
    offset: float | None = None,
    fade_in_start: float | None = None,
    fade_in_stop: float | None = None,
    fade_out_start: float | None = None,
    fade_out_stop: float | None = None,
    color_index: int | None = None,
) -> None:
    """Change the current project's soundtrack settings."""
    optional_args = [
        int(mute) if mute is not None else None,
        volume,
        offset,
        (fade_in_start, fade_in_stop, fade_out_start, fade_out_stop),
        color_index,
    ]

    args = validate_args_list(optional_args)

    send_cmd(
        "tv_SoundProjectAdjust",
        track_index,
        *args,
        error_values=[-2, -3],
    )


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid project id",
)
def tv_project_header_info_get(project_id: str) -> str:
    """Get the project header info.

    Raises:
        NoObjectWithIdError: if given an invalid project id
    """
    return send_cmd(
        "tv_ProjectHeaderInfo",
        project_id,
        error_values=[GrgErrorValue.ERROR],
    ).strip('"')


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid project id",
)
def tv_project_header_info_set(project_id: str, text: str) -> None:
    """Set the project header info.

    Raises:
        NoObjectWithIdError: if given an invalid project id
    """
    send_cmd(
        "tv_ProjectHeaderInfo",
        project_id,
        text,
        error_values=[GrgErrorValue.ERROR],
    )


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid project id",
)
def tv_project_header_author_get(project_id: str) -> str:
    """Get the project author info.

    Raises:
        NoObjectWithIdError: if given an invalid project id
    """
    return send_cmd(
        "tv_ProjectHeaderAuthor",
        project_id,
        error_values=[GrgErrorValue.ERROR],
    ).strip('"')


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid project id",
)
def tv_project_header_author_set(project_id: str, text: str) -> None:
    """Set the project author info.

    Raises:
        NoObjectWithIdError: if given an invalid project id
    """
    send_cmd(
        "tv_ProjectHeaderAuthor",
        project_id,
        text,
        error_values=[GrgErrorValue.ERROR],
    )


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid project id",
)
def tv_project_header_notes_get(project_id: str) -> str:
    """Get the project notes.

    Raises:
        NoObjectWithIdError: if given an invalid project id
    """
    return send_cmd(
        "tv_ProjectHeaderNotes",
        project_id,
        error_values=[GrgErrorValue.ERROR],
    ).strip('"')


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid project id",
)
def tv_project_header_notes_set(project_id: str, text: str) -> None:
    """Set the project notes.

    Raises:
        NoObjectWithIdError: if given an invalid project id
    """
    send_cmd(
        "tv_ProjectHeaderNotes",
        project_id,
        text,
        error_values=[GrgErrorValue.ERROR],
    )


def tv_start_frame_get() -> int:
    """Get the start frame of the current project."""
    return int(send_cmd("tv_StartFrame"))


def tv_start_frame_set(start_frame: int) -> int:
    """Set the start frame of the current project."""
    return int(send_cmd("tv_StartFrame", start_frame))



"""Scene related George functions."""

from __future__ import annotations

from pytvpaint.george.client import send_cmd, try_cmd
from pytvpaint.george.grg_base import GrgErrorValue


@try_cmd(exception_msg="No scene at provided position")
def tv_scene_enum_id(position: int) -> int:
    """Get the id of the scene at the given position in the current project.

    Raises:
        GeorgeError: if no scene found at the provided position
    """
    return int(send_cmd("tv_SceneEnumId", position, error_values=[GrgErrorValue.NONE]))


def tv_scene_current_id() -> int:
    """Get the id of the current scene."""
    return int(send_cmd("tv_SceneCurrentId"))


def tv_scene_move(scene_id: int, position: int) -> None:
    """Move a scene to another position."""
    send_cmd("tv_SceneMove", scene_id, position)


def tv_scene_new() -> None:
    """Create a new scene (with a new clip) after the current scene."""
    send_cmd("tv_SceneNew")


def tv_scene_duplicate(scene_id: int) -> None:
    """Duplicate the given scene."""
    send_cmd("tv_SceneDuplicate", scene_id)


def tv_scene_close(scene_id: int) -> None:
    """Remove the given scene."""
    send_cmd("tv_SceneClose", scene_id)



"""PyTVPaint package logger."""

from __future__ import annotations

import logging
import os
import sys
from logging.handlers import RotatingFileHandler
from pathlib import Path


def _get_logger() -> logging.Logger:
    log_format = "[%(asctime)s] %(name)s / %(levelname)s -- %(message)s"
    formatter = logging.Formatter(log_format)

    logger_name = "pytvpaint"
    logger = logging.getLogger(logger_name)
    log_level = logging.getLevelName(os.getenv("PYTVPAINT_LOG_LEVEL", "INFO").upper())

    logger.setLevel(log_level)
    stream_handler = logging.StreamHandler(sys.stdout)
    stream_handler.setFormatter(formatter)
    logger.addHandler(stream_handler)

    log_path = os.getenv("PYTVPAINT_LOG_PATH", "")
    if log_path:
        log_path_obj = Path(log_path)
        log_path_obj.parent.mkdir(parents=True, exist_ok=True)

        file_handler = RotatingFileHandler(
            filename=log_path_obj.as_posix(),
            mode="a+",
            maxBytes=1024 * 1024,
            backupCount=5,
        )

        file_handler.set_name(logger_name)
        file_handler.setFormatter(formatter)
        file_handler.setLevel(log_level)
        logger.addHandler(file_handler)

    return logger


log = _get_logger()



"""Clip related George functions."""

from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path
from typing import Any

from pytvpaint.george.client import send_cmd, try_cmd
from pytvpaint.george.client.parse import (
    args_dict_to_list,
    tv_parse_dict,
    tv_parse_list,
    validate_args_list,
)
from pytvpaint.george.exceptions import NoObjectWithIdError
from pytvpaint.george.grg_base import (
    FieldOrder,
    GrgErrorValue,
    SaveFormat,
    SpriteLayout,
    TVPSound,
)


@dataclass(frozen=True)
class TVPClip:
    """TVPaint clip info values."""

    id: int = field(metadata={"parsed": False})

    name: str
    is_current: bool
    is_hidden: bool
    is_selected: bool
    storyboard_start_frame: int
    first_frame: int
    last_frame: int
    frame_count: int
    mark_in: int
    mark_out: int
    color_idx: int


class PSDSaveMode(Enum):
    """PSD save modes.

    Attributes:
        ALL:
        IMAGE:
        MARKIN:
    """

    ALL = "all"
    IMAGE = "image"
    MARKIN = "markin"


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid clip id",
)
def tv_clip_info(clip_id: int) -> TVPClip:
    """Get the information of the given clip.

    Raises:
        NoObjectWithIdError: if given an invalid clip id
    """
    result = send_cmd("tv_ClipInfo", clip_id, error_values=[GrgErrorValue.EMPTY])
    clip = tv_parse_dict(result, with_fields=TVPClip)
    clip["id"] = clip_id
    return TVPClip(**clip)


@try_cmd(
    exception_msg="Invalid scene id or clip position or elements have been removed"
)
def tv_clip_enum_id(scene_id: int, clip_position: int) -> int:
    """Get the id of the clip at the given position inside the given scene.

    Raises:
        GeorgeError: if given an invalid scene id or clip position or elements have been removed
    """
    return int(
        send_cmd(
            "tv_ClipEnumId",
            scene_id,
            clip_position,
            error_values=[GrgErrorValue.NONE],
        )
    )


def tv_clip_current_id() -> int:
    """Get the id of the current clip."""
    return int(send_cmd("tv_ClipCurrentId"))


def tv_clip_new(name: str) -> None:
    """Create a new clip."""
    send_cmd("tv_ClipNew", name, handle_string=False)


def tv_clip_duplicate(clip_id: int) -> None:
    """Duplicate the given clip."""
    send_cmd("tv_ClipDuplicate", clip_id)


def tv_clip_close(clip_id: int) -> None:
    """Remove the given clip."""
    send_cmd("tv_ClipClose", clip_id)


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid clip id",
)
def tv_clip_name_get(clip_id: int) -> str:
    """Get the clip name.

    Raises:
        NoObjectWithIdError: if given an invalid clip id
    """
    return send_cmd("tv_ClipName", clip_id, error_values=[GrgErrorValue.EMPTY])


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid clip id",
)
def tv_clip_name_set(clip_id: int, name: str) -> None:
    """Set the clip name.

    Raises:
        NoObjectWithIdError: if given an invalid clip id
    """
    send_cmd("tv_ClipName", clip_id, name, error_values=[GrgErrorValue.EMPTY])


def tv_clip_move(clip_id: int, scene_id: int, position: int) -> None:
    """Manage clip position."""
    send_cmd("tv_ClipMove", clip_id, scene_id, position)


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid clip id",
)
def tv_clip_hidden_get(clip_id: int) -> bool:
    """Get clip visibility.

    Raises:
        NoObjectWithIdError: if given an invalid clip id
    """
    res = send_cmd("tv_ClipHidden", clip_id, error_values=[GrgErrorValue.EMPTY])
    return bool(int(res))


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid clip id",
)
def tv_clip_hidden_set(clip_id: int, new_state: bool) -> None:
    """Set clip visibility.

    Raises:
        NoObjectWithIdError: if given an invalid clip id
    """
    send_cmd(
        "tv_ClipHidden", clip_id, int(new_state), error_values=[GrgErrorValue.EMPTY]
    )


def tv_clip_select(clip_id: int) -> None:
    """Activate/Make current the given clip."""
    send_cmd("tv_ClipSelect", clip_id)


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid clip id",
)
def tv_clip_selection_get(clip_id: int) -> bool:
    """Get the clip's selection state.

    Raises:
        NoObjectWithIdError: if given an invalid clip id
    """
    res = send_cmd("tv_ClipSelection", clip_id, error_values=[-1])
    return bool(int(res))


@try_cmd(
    raise_exc=NoObjectWithIdError,
    exception_msg="Invalid clip id",
)
def tv_clip_selection_set(clip_id: int, new_state: bool) -> None:
    """Set the clip's selection state.

    Raises:
        NoObjectWithIdError: if given an invalid clip id
    """
    send_cmd("tv_ClipSelection", clip_id, int(new_state), error_values=[-1])


def tv_first_image() -> int:
    """Get the first image of the clip."""
    return int(send_cmd("tv_FirstImage"))


def tv_last_image() -> int:
    """Get the last image of the clip."""
    return int(send_cmd("tv_LastImage"))


@try_cmd(exception_msg="Invalid format for sequence")
def tv_load_sequence(
    seq_path: Path | str,
    offset_count: tuple[int, int] | None = None,
    field_order: FieldOrder | None = None,
    stretch: bool | None = None,
    time_stretch: bool | None = None,
    preload: bool | None = None,
) -> int:
    """Load a sequence of images or movie in a new layer.

    Args:
        seq_path: the first file of the sequence to load
        offset_count: the start and number of image of sequence to load. Defaults to None.
        field_order: the field order. Defaults to None.
        stretch: Stretch each image to the size of the layer. Defaults to None.
        time_stretch: Once loaded, the layer will have a new number of image corresponding to the project framerate. Defaults to None.
        preload: Load all the images in memory, no more reference on the files. Defaults to None.

    Raises:
        FileNotFoundError: if the sequence file doesn't exist
        GeorgeError: if the input file is in an invalid format

    Returns:
        the number of images of the new layer
    """
    seq_path = Path(seq_path)

    if not seq_path.exists():
        raise FileNotFoundError(f"File not found at: {seq_path.as_posix()}")

    args: list[int | str | None] = [field_order.value if field_order else None]

    if offset_count is not None:
        offset, count = offset_count
        args.insert(0, count)
        args.insert(0, offset)

    # Filter None inline arguments
    args = [a for a in args if a is not None]

    args += args_dict_to_list(
        {
            "stretch": stretch,
            "timestretch": time_stretch,
            "preload": preload,
        }
    )

    result = send_cmd(
        "tv_LoadSequence",
        seq_path.as_posix(),
        *args,
        error_values=[-1],
    )

    return int(result)


def tv_save_sequence(
    export_path: Path | str,
    mark_in: int | None = None,
    mark_out: int | None = None,
) -> None:
    """Save the current clip.

    Raises:
        NotADirectoryError: if the export directory doesn't exist
    """
    export_path = Path(export_path).resolve()

    if not export_path.parent.exists():
        raise NotADirectoryError(
            "Can't save the sequence because parent"
            f"folder does not exist: {export_path.parent.as_posix()}"
        )

    args: list[Any] = [export_path.as_posix()]

    if mark_in and mark_out:
        args.extend([mark_in, mark_out])

    send_cmd("tv_SaveSequence", *args)


@try_cmd(exception_msg="No bookmark at provided position")
def tv_bookmarks_enum(position: int) -> int:
    """Get the frame (in the clip) corresponding to the bookmark at the given position.

    Raises:
        GeorgeError: if no bookmark found at provided position
    """
    return int(
        send_cmd("tv_BookmarksEnum", position, error_values=[GrgErrorValue.NONE])
    )


def tv_bookmark_set(frame: int) -> None:
    """Set a bookmark at the given frame."""
    send_cmd("tv_BookmarkSet", frame)


def tv_bookmark_clear(frame: int) -> None:
    """Remove a bookmark at the given frame."""
    send_cmd("tv_BookmarkClear", frame)


def tv_bookmark_next() -> None:
    """Go to the next bookmarked frame."""
    send_cmd("tv_BookmarkNext")


def tv_bookmark_prev() -> None:
    """Go to the previous bookmarked frame."""
    send_cmd("tv_BookmarkPrev")


def tv_clip_color_get(clip_id: int) -> int:
    """Get the clip color."""
    return int(send_cmd("tv_ClipColor", clip_id, error_values=[GrgErrorValue.EMPTY]))


def tv_clip_color_set(clip_id: int, color_index: int) -> None:
    """Set the clip color."""
    send_cmd("tv_ClipColor", clip_id, color_index, error_values=[GrgErrorValue.EMPTY])


def tv_clip_action_get(clip_id: int) -> str:
    """Get the action text of the clip."""
    # We explicitly check if the clip exists because the error value is an empty string, and we can't determine if the
    # action text is empty or the clip_id is invalid...
    tv_clip_name_get(clip_id)
    return send_cmd("tv_ClipAction", clip_id)


def tv_clip_action_set(clip_id: int, text: str) -> None:
    """Set the action text of the clip."""
    # See tv_clip_action_get above
    tv_clip_name_get(clip_id)
    send_cmd("tv_ClipAction", clip_id, text)


def tv_clip_dialog_get(clip_id: int) -> str:
    """Get the dialog text of the clip."""
    # See tv_clip_action_get above
    tv_clip_name_get(clip_id)
    return send_cmd("tv_ClipDialog", clip_id)


def tv_clip_dialog_set(clip_id: int, dialog: str) -> None:
    """Set the dialog text of the clip."""
    # See tv_clip_action_get above
    tv_clip_name_get(clip_id)
    send_cmd("tv_ClipDialog", clip_id, dialog)


def tv_clip_note_get(clip_id: int) -> str:
    """Get the note text of the clip."""
    # See tv_clip_action_get above
    tv_clip_name_get(clip_id)
    return send_cmd("tv_ClipNote", clip_id)


def tv_clip_note_set(clip_id: int, note: str) -> None:
    """Set the note text of the clip."""
    # See tv_clip_action_get above
    tv_clip_name_get(clip_id)
    send_cmd("tv_ClipNote", clip_id, note)


@try_cmd(exception_msg="Can't create file")
def tv_save_clip(export_path: Path | str) -> None:
    """Save the current clip in .tvp format.

    Raises:
        GeorgeError: if file couldn't be saved
    """
    export_path = Path(export_path)
    send_cmd("tv_SaveClip", export_path.as_posix())


def tv_save_display(export_path: Path | str) -> None:
    """Save the display."""
    export_path = Path(export_path).resolve()
    send_cmd("tv_SaveDisplay", export_path.as_posix())


def tv_clip_save_structure_json(
    export_path: Path | str,
    file_format: SaveFormat,
    fill_background: bool = False,
    folder_pattern: str | None = None,
    file_pattern: str | None = None,
    visible_layers_only: bool = True,
    all_images: bool = False,
    ignore_duplicates: bool = False,
    exclude_instance_names: list[str] | None = None,
) -> None:
    """Save the current clip structure in json.

    Args:
        export_path: the JSON export path
        file_format: file format to use for rendering
        fill_background: add a background. Defaults to None.
        folder_pattern: the folder name pattern (%li: layer index, %ln: layer name, %fi: file index (added in 11.0.8)). Defaults to None.
        file_pattern: the file name pattern (%li: layer index, %ln: layer name, %ii: image index, %in: image name, %fi: file index (added in 11.0.8)). Defaults to None.
        visible_layers_only: export only visible layers. Defaults to None.
        all_images: export all images. Defaults to None.
        ignore_duplicates: Ignore duplicates images. Defaults to None.
        exclude_instance_names: the instances names which won't be processed/exported. Defaults to None.

    Raises:
        ValueError: the parent folder doesn't exist
    """
    export_path = Path(export_path).resolve()

    if not export_path.parent.exists():
        raise ValueError(
            "Can't write file because the destination folder doesn't exist"
        )

    args = [export_path.as_posix(), "JSON"]

    dict_args = {
        "fileformat": file_format.value,
        "background": int(fill_background) if fill_background else None,
        "patternfolder": folder_pattern,
        "patternfile": file_pattern,
        "onlyvisiblelayers": int(visible_layers_only),
        "allimages": int(all_images),
        "ignoreduplicateimages": int(ignore_duplicates),
        "excludenames": (
            ";".join(exclude_instance_names) if exclude_instance_names else None
        ),
    }
    args.extend(args_dict_to_list(dict_args))

    send_cmd("tv_ClipSaveStructure", *args, error_values=[-1])


def tv_clip_save_structure_psd(
    export_path: Path | str,
    mode: PSDSaveMode,
    image: int | None = None,
    mark_in: int | None = None,
    mark_out: int | None = None,
) -> None:
    """Save the current clip as a PSD.

    Args:
        export_path: _description_
        mode: all will export all layers to a single PSD
        image: will only export the given image. Defaults to None.
        mark_in: start frame to render. Defaults to None.
        mark_out: end frame to render. Defaults to None.
    """
    export_path = Path(export_path)

    if not export_path.parent.exists():
        raise ValueError(
            "Can't write file because the destination folder doesn't exist"
        )

    args_dict: dict[str, str | int | None]

    if mode == PSDSaveMode.ALL:
        args_dict = {"mode": "all"}
    elif mode == PSDSaveMode.IMAGE:
        if image is None:
            raise ValueError("Image must be defined")
        args_dict = {"image": image}
    else:  # Markin
        if mark_in is None or mark_out is None:
            raise ValueError("mark_in and mark_out must be defined")
        args_dict = {"markin": mark_in, "markout": mark_out}

    args = args_dict_to_list(args_dict)

    send_cmd(
        "tv_ClipSaveStructure",
        export_path.as_posix(),
        "PSD",
        *args,
        error_values=[-1],
    )


def tv_clip_save_structure_csv(
    export_path: Path | str,
    all_images: bool | None = None,
    exposure_label: str | None = None,
) -> None:
    """Save the current clip as a CSV.

    Args:
        export_path: the .csv export path
        all_images: export all images or only instances. Defaults to None.
        exposure_label: give a label when the image is an exposure. Defaults to None.
    """
    export_path = Path(export_path)

    args = args_dict_to_list(
        {
            "allimages": int(bool(all_images)),
            "exposurelabel": exposure_label,
        }
    )

    send_cmd(
        "tv_ClipSaveStructure",
        export_path.as_posix(),
        "CSV",
        *args,
        error_values=[-1],
    )


def tv_clip_save_structure_sprite(
    export_path: Path | str,
    layout: SpriteLayout | None = None,
    space: int | None = None,
) -> None:
    """Save the current clip as sprites in one image.

    Args:
        export_path: the export path of the sprite image
        layout: the sprite layout. Defaults to None.
        space: the space between each sprite in the image. Defaults to None.
    """
    export_path = Path(export_path)

    args = args_dict_to_list(
        {
            "layout": layout.value if layout is not None else None,
            "space": space,
        }
    )

    send_cmd(
        "tv_ClipSaveStructure",
        export_path.as_posix(),
        "sprite",
        *args,
        error_values=[-1],
    )


def tv_clip_save_structure_flix(
    export_path: Path | str,
    mark_in: int | None = None,
    mark_out: int | None = None,
    parameters_import: str | None = None,
    parameters_file: str | None = None,
    send: bool | None = None,
    original_file: str | Path | None = None,
) -> None:
    """Save the current clip for Flix.

    Args:
        export_path: the .xml export path
        mark_in: the start frame to render. Defaults to None.
        mark_out: the end frame to render. Defaults to None.
        parameters_import: the attribute(s) of the global <flixImport> tag (waitForSource/...). Defaults to None.
        parameters_file: the attribute(s) of each <image> (file) tag (dialogue/...). Defaults to None.
        send: open a browser with the prefilled url. Defaults to None.
        original_file: the original reference tvpp file path. Defaults to None.
    """
    export_path = Path(export_path)

    args_dict = {
        "markin": mark_in,
        "markout": mark_out,
        "parametersimport": parameters_import,
        "parametersfile": parameters_file,
        "send": int(send) if send else None,
        "originalfile": Path(original_file).as_posix() if original_file else None,
    }

    args = args_dict_to_list(args_dict)

    send_cmd(
        "tv_ClipSaveStructure",
        export_path.as_posix(),
        "Flix",
        *args,
        error_values=[-1],
    )


def tv_sound_clip_info(clip_id: int, track_index: int) -> TVPSound:
    """Get information about a soundtrack."""
    res = send_cmd("tv_SoundClipInfo", clip_id, track_index, error_values=[-1, -2, -3])
    res_parse = tv_parse_list(res, with_fields=TVPSound)
    return TVPSound(**res_parse)


def tv_sound_clip_new(sound_path: Path | str) -> None:
    """Add a new soundtrack."""
    path = Path(sound_path)
    if not path.exists():
        raise ValueError(f"Sound file not found at : {path.as_posix()}")
    send_cmd("tv_SoundClipNew", path.as_posix(), error_values=[-1, -2, -3, -4])


def tv_sound_clip_remove(track_index: int) -> None:
    """Remove a soundtrack."""
    send_cmd("tv_SoundClipRemove", track_index, error_values=[-2])


def tv_sound_clip_reload(clip_id: int, track_index: int) -> None:
    """Reload a soundtrack from its file.

    Args:
        clip_id: the clip id (only works with `0` being the current clip)
        track_index: the sound clip track index

    Warning:
        this doesn't accept a proper clip id, only `0` seem to work for the current clip
    """
    send_cmd("tv_SoundClipReload", clip_id, track_index, error_values=[-1, -2, -3])


def tv_sound_clip_adjust(
    track_index: int,
    mute: bool | None = None,
    volume: float | None = None,
    offset: float | None = None,
    fade_in_start: float | None = None,
    fade_in_stop: float | None = None,
    fade_out_start: float | None = None,
    fade_out_stop: float | None = None,
    color_index: int | None = None,
) -> None:
    """Change a soundtracks settings."""
    optional_args = [
        int(mute) if mute is not None else None,
        volume,
        offset,
        (fade_in_start, fade_in_stop, fade_out_start, fade_out_stop),
        color_index,
    ]

    args = validate_args_list(optional_args)
    send_cmd("tv_SoundClipAdjust", track_index, *args, error_values=[-2, -3])


def tv_layer_image_get() -> int:
    """Get the current frame of the current clip."""
    return int(send_cmd("tv_LayerGetImage"))


def tv_layer_image(frame: int) -> None:
    """Set the current frame of the current clip."""
    send_cmd("tv_LayerImage", frame)
